Set Universe Polymorphism.
Require Import Ltac2.Ltac2.

(* basic union find with extra logic for keeping track of predicesors and proofs *)
Require UFProof.


(* note I skip mutablility here in an effort to keep it simpler *)

Ltac2 Type ('a, 'h) Node :=
  [ Repr ( int, 'h ) (* class *)
  | Node ( 'a UFProof.t, int ) (* proof, parent *)
  ].

Ltac2 Type ('a,'h) t := (('a,'h) Node) array.

(*
  find the class that a term belongs to, optimizing the structure along the way
  
  returns a tuple of (repTerm, size, info, proof that term = repTerm)
*)
Ltac2 rec findClass ( uf : ('a, 'h) t ) ( term: int ) :=
  match (Array.get uf term) with
  | Repr size info => ( term, size, info, UFProof.refl term )
  | Node proof parent =>
        let ( repTerm, size, info, prf) := findClass uf parent in
        let newProof := UFProof.trans term proof parent prf repTerm in
          Array.set uf term (Node newProof repTerm) ;
          (repTerm,  size, info,  newProof)
  end.


(*
  unfies two terms, returing the *origional* left and right parents of each 
*)
Ltac2 union 
  (combine: 'h -> 'h -> 'h)
  (uf : ('a, 'h) t ) (lhs: int) (rhs: int) (proof: 'a UFProof.t) :=
  let (lhsRep, lhsSize, lhsInfo, lhsPrf) := findClass uf lhs in 
  let (rhsRep, rhsSize, rhsInfo, rhsPrf) := findClass uf rhs in 
  match (Int.equal lhsRep rhsRep) with
  | false =>
          let classEqProof := (UFProof.transportEq lhs lhsRep rhs rhsRep lhsPrf rhsPrf proof) in
          match (Int.ge lhsSize rhsSize) with
          | true => 
              (* lhs is larger so it should be the new root *)
              Array.set uf lhsRep (Repr 
                (Int.add lhsSize rhsSize)
                (combine lhsInfo rhsInfo)); 
              Array.set uf rhsRep (Node 
                (* make a proof that rhsRep = lhsRep *)
                (UFProof.transportEq rhs rhsRep  lhs lhsRep  rhsPrf lhsPrf (UFProof.sym lhs rhs proof))
                lhsRep)
          | false =>
              Array.set uf lhsRep (Node 
                        (* make a proof that rhsRep = lhsRep *)
                        (UFProof.transportEq lhs lhsRep rhs rhsRep lhsPrf rhsPrf proof)
                        rhsRep);
              Array.set uf rhsRep (Repr
                (Int.add lhsSize rhsSize)
                (combine lhsInfo rhsInfo))
          end;
          Some (
            lhsInfo,
            rhsInfo
          )
  | true => None
  end.

(* returns a boolean of if two terms are congurent *)
Ltac2 isCongruent (uf : ('a,'h) t ) (lhs: int) (rhs: int) :=
  let (lhsRep, _, _, lhsPrf) := findClass uf lhs in 
  let (rhsRep, _, _, rhsPrf) := findClass uf rhs in
  Int.equal lhsRep rhsRep.

(* maybe returns a proof if two terms are congruent *)
Ltac2 congruent (uf : ('a,'h) t ) (lhs: int) (rhs: int) := 
  let (lhsRep, _, _, lhsPrf) := findClass uf lhs in 
  let (rhsRep, _, _, rhsPrf) := findClass uf rhs in 
  match (Int.equal lhsRep rhsRep) with
  | false => None
  | true => Some (UFProof.trans lhs lhsPrf lhsRep (UFProof.sym rhs lhsRep rhsPrf) rhs)
  end.


Ltac2 fromInfo (x : 'h array) := Array.map (fun i => Repr 1 i) x.
